import numpy as np
import pandas as pd

import unittest

from tsforest.forecaster_lightgbm import LightGBMForecaster
from tsforest.utils import make_time_range

DATA_PATH = './tests/tests_data/data.csv'

class TestLightGMB(unittest.TestCase):

    def test_it_fit(self):
        print("##################"*12)
        data = pd.read_csv(DATA_PATH, parse_dates=['ds'])

        fcaster = LightGBMForecaster(model_params={'num_iterations':30,
                                                   'learning_rate':0.3}, 
                                     features=['calendar', 'calendar_cyclical'])
        fcaster.fit(train_data=data)
    
    def test_it_fit_with_valid_period(self):
        data = pd.read_csv(DATA_PATH, parse_dates=['ds'])
        valid_period = make_time_range('2019-06-01', '2019-06-30', freq='D')

        fcaster = LightGBMForecaster(model_params={'num_iterations':30,
                                                   'learning_rate':0.3}, 
                                     features=['calendar', 'calendar_cyclical'])
        fcaster.fit(train_data=data, valid_period=valid_period)

    def test_it_fit_with_lag_features(self):
        data = pd.read_csv(DATA_PATH, parse_dates=['ds'])
        valid_period = make_time_range('2019-06-01', '2019-06-30', freq='D')

        fcaster = LightGBMForecaster(model_params={'num_iterations':30,
                                                   'learning_rate':0.3}, 
                                     features=['calendar', 'calendar_cyclical','lag'],
                                     lags=[1,2,3,4,5,6,7])
        fcaster.fit(train_data=data, valid_period=valid_period)

    def test_it_fit_with_rw_features(self):
        data = pd.read_csv(DATA_PATH, parse_dates=['ds'])
        valid_period = make_time_range('2019-06-01', '2019-06-30', freq='D')

        fcaster = LightGBMForecaster(model_params={'num_iterations':30,
                                                   'learning_rate':0.3}, 
                                     features=['calendar', 'calendar_cyclical','rw'],
                                     window_functions=['mean','median','min','max','sum'],
                                     window_sizes=[7,14,21,28])
        fcaster.fit(train_data=data, valid_period=valid_period)

    def test_it_fit_predict(self):
        data = pd.read_csv(DATA_PATH, parse_dates=['ds'])
        test_period = make_time_range('2019-07-01', '2019-07-31', freq='D')

        fcaster = LightGBMForecaster(model_params={'num_iterations':30,
                                                   'learning_rate':0.3}, 
                                     features=['calendar', 'calendar_cyclical'])
        fcaster.fit(train_data=data)
        prediction_dataframe = fcaster.predict(test_period)

    def test_it_fit_predict_with_lag_features(self):
        data = pd.read_csv(DATA_PATH, parse_dates=['ds'])
        test_period = make_time_range('2019-07-01', '2019-07-31', freq='D')

        fcaster = LightGBMForecaster(model_params={'num_iterations':30,
                                                   'learning_rate':0.3}, 
                                     features=['calendar', 'calendar_cyclical','lag'],
                                     lags=[1,2,3,4,5,6,7])
        fcaster.fit(train_data=data)
        prediction_dataframe = fcaster.predict(test_period)

    def test_it_fit_predict_with_rw_features(self):
        data = pd.read_csv(DATA_PATH, parse_dates=['ds'])
        test_period = make_time_range('2019-07-01', '2019-07-31', freq='D')

        fcaster = LightGBMForecaster(model_params={'num_iterations':30,
                                                   'learning_rate':0.3}, 
                                     features=['calendar', 'calendar_cyclical','rw'],
                                     window_functions=['mean','median','min','max','sum'],
                                     window_sizes=[7,14,21,28])
        fcaster.fit(train_data=data)
        prediction_dataframe = fcaster.predict(test_period)

    def test_it_fit_evaluate(self):
        data = pd.read_csv(DATA_PATH, parse_dates=['ds'])
        train_data = data.query('ds < "2019-06-01"')
        eval_data = data.query('ds >= "2019-06-01"')

        fcaster = LightGBMForecaster(model_params={'num_iterations':30,
                                                   'learning_rate':0.3}, 
                                     features=['calendar', 'calendar_cyclical'])
        fcaster.fit(train_data=train_data)
        error = fcaster.evaluate(eval_data)

        assert type(error)==np.float64, \
            f'fcaster.evaluate returns {error} which is not of type numpy.float64'



import numpy as np
import pandas as pd
pd.options.mode.chained_assignment = None
import calendar
import h2o

from tsforest.config import (calendar_sequential_features_types,
                             calendar_cyclical_features_types,
                             all_features_types)

def fill_time_gaps(data):
    """
    Parameters
    ----------
    data: pandas.DataFrame
        dataframe with columns 'ds' (dtype datetime64) and 'y' 
    """
    assert set(['ds','y']) <= set(data.columns), \
        'data must contain the column "ds"'
    filled_data = (data
                   .resample("D", on='ds').y.mean()
                   .interpolate("linear")
                   .reset_index())
    filled_data = pd.merge(filled_data, data.drop('y', axis=1), on=['ds'], how='left')
    return filled_data

def compute_calendar_features(start_time, end_time, freq='D'):
    """
    Parameters
    ----------
    start_time: str
        start time for time range
    end_time: str
        end time for time range
    freq: str
        frequency string of pandas.date_range function
    """
    date_range = pd.date_range(start=start_time, end=end_time, freq=freq)
    
    calendar_data = pd.DataFrame()
    calendar_data['ds'] = date_range
    calendar_data['week_day'] = date_range.dayofweek
    calendar_data['month_day'] = date_range.day
    calendar_data['year_day'] = date_range.dayofyear
    calendar_data['year_week'] = date_range.weekofyear
    calendar_data['month'] = date_range.month
    calendar_data['year'] = date_range.year
    days_of_month = calendar_data.apply(lambda x: calendar.monthrange(x.year,x.month)[1], axis=1)
    calendar_data['month_progress'] = calendar_data.month_day/days_of_month

    # cyclical encodings: day of week
    calendar_data['week_day_cos'] = np.cos(calendar_data.week_day*(2.*np.pi/7))
    calendar_data['week_day_sin'] = np.sin(calendar_data.week_day*(2.*np.pi/7))
    # cyclical encodings: day of year
    calendar_data['year_day_cos'] = np.cos((calendar_data.year_day-1)*(2.*np.pi/366))
    calendar_data['year_day_sin'] = np.sin((calendar_data.year_day-1)*(2.*np.pi/366))
    # cyclical encodings: week of year
    calendar_data['year_week_cos'] = np.cos((calendar_data.year_week-1)*(2.*np.pi/52))
    calendar_data['year_week_sin'] = np.sin((calendar_data.year_week-1)*(2.*np.pi/52))
    # cyclical encodings: month of year
    calendar_data['month_cos'] = np.cos((calendar_data.month-1)*(2.*np.pi/12))
    calendar_data['month_sin'] = np.sin((calendar_data.month-1)*(2.*np.pi/12))
    # week_day shifted to 1-7
    calendar_data['week_day'] += 1
    return calendar_data

def compute_lag_features(data, lags):
    """
    data: pandas.Dataframe
        dataframe with column 'y' (response values)
    lags: list
        list of integer lag values
    """
    assert 'y' in data.columns, \
        'Missing "y" column in dataframe'
    features = [data.y.shift(lag) for lag in lags]
    features_names = [f'lag_{lag}' for lag in lags]
    lag_features = pd.concat([data.ds] + features, axis=1)
    lag_features.columns = ['ds'] + features_names
    return lag_features  

def compute_rw_features(data, window_sizes, window_functions):
    """
    data: pandas.Dataframe
        dataframe with column 'y' (response values)
    window_sizes: list
        list of integer window sizes values
    window_functions: list
        list of string names of the window functions
    """
    assert 'y' in data.columns, \
        'Missing "y" column in dataframe'
    # assert window functions in availabe funcs...
    shifted = data.y.shift(1)    
    features = [getattr(shifted.rolling(window), window_func).__call__()
                for window_func in window_functions
                for window in window_sizes]
    features_names = [f'{window_func}_{window}'
                      for window_func in window_functions
                      for window in window_sizes]
    rw_features = pd.concat([data.ds] + features, axis=1)
    rw_features.columns = ['ds'] + features_names
    return rw_features

class FeaturesGenerator():

    def __init__(self, include_features, lags=None, window_sizes=None, 
                 window_functions=None):
        """
        Parameters
        ----------
        include_features: list
            Features to included in computation
        lags: list
            List of integer lag values
        window_sizes: list
            List of integer window sizes values
        window_functions: list
            List of string names of the window functions
        """
        self.include_features = include_features
        self.lags = lags
        self.window_sizes = window_sizes
        self.window_functions = window_functions

    def compute_train_features(self, data):
        """
        Parameters
        ----------
        data : pd.DataFrame
            Dataframe with (at least) columns: 'ds' and 'y'
        Returns
        ----------
        all_features: pd.Dataframe
            Dataframe containing all the features for the time series
        """
        self.train_data = data
        # in case of time gaps
        filled_data = fill_time_gaps(data)

        # list with all the dataframes of features
        all_features_list = list()

        # generating the calendar features
        if {'calendar','calendar_cyclical'} & set(self.include_features):
            calendar_features = (compute_calendar_features(data.ds.min(), data.ds.max())
                                 .merge(data.loc[:, ['ds']], how='inner', on=['ds']))
        if "calendar" not in self.include_features:
            columns_to_drop = list(calendar_sequential_features_types.keys())
            calendar_features.drop(columns=columns_to_drop, inplace=True)
        elif "calendar_cyclical" not in self.include_features:
            columns_to_drop = list(calendar_cyclical_features_types.keys())
            calendar_features.drop(columns=columns_to_drop, inplace=True)
        all_features_list.append(calendar_features.set_index('ds'))
                
        if "lag" in self.include_features:
            lag_features = (compute_lag_features(filled_data, lags=self.lags)
                            .merge(data.loc[:, ['ds']], how='inner', on=['ds']))
            all_features_list.append(lag_features.set_index('ds'))

        if "rw" in self.include_features:
            rw_features = (compute_rw_features(filled_data, 
                                               window_sizes=self.window_sizes, 
                                               window_functions=self.window_functions)
                           .merge(data.loc[:, ['ds']], how='inner', on=['ds']))
            all_features_list.append(rw_features.set_index('ds'))

        # merging all features
        all_features_list.append(data.set_index("ds"))
        all_features = (pd.concat(all_features_list, axis=1)
                        .reset_index()
                        .rename({'index':'ds'}, axis=1)
                        .assign(y = lambda x: x['y'].fillna(0.)))
        categorical_features = [feature for feature,dtype in all_features_types.items()
                                if feature in all_features.columns and dtype=='categorical']
        return all_features,categorical_features
        
    def compute_test_features(self, data):
        """
        Parameters
        ----------
        data: pandas.DataFrame
            Dataframe with (at least) column: 'ds'
        """
        self.test_data = data

        # list with all the dataframes of features
        all_features_list = list()

        # generating the calendar features
        if np.any(['calendar' in feat for feat in self.include_features]):
            calendar_features = (compute_calendar_features(data.ds.min(), data.ds.max())
                                 .merge(data.loc[:, ['ds']], how='inner', on=['ds']))
        if "calendar" not in self.include_features:
            columns_to_drop = list(calendar_sequential_features_types.keys())
            calendar_features.drop(columns=columns_to_drop, inplace=True)
        elif "calendar_cyclical" not in self.include_features:
            columns_to_drop = list(calendar_cyclical_features_types.keys())
            calendar_features.drop(columns=columns_to_drop, inplace=True)
        all_features_list.append(calendar_features.set_index('ds'))

        if "lag" in self.include_features:
            column_names = [f'lag_{lag}' for lag in self.lags]
            lag_features = pd.DataFrame(np.nan, index=data.ds, 
                                        columns=column_names)
            all_features_list.append(lag_features)
    
        if "rw" in self.include_features:
            column_names = [f'{window_func}_{window}'
                            for window_func in self.window_functions
                            for window in self.window_sizes]
            rw_features = pd.DataFrame(np.nan, index=data.ds,
                                       columns=column_names)
            all_features_list.append(rw_features)
        
        # merging all features
        all_features_list.append(data.set_index("ds"))
        all_features = pd.concat(all_features_list, axis=1)
        all_features.reset_index(inplace=True)
        categorical_features = [feature for feature,dtype in all_features_types.items()
                                if feature in all_features.columns and dtype=='categorical']
        return all_features,categorical_features



# h2o gbm default parameters
gbm_parameters = {
    'ntrees':300,
    'max_depth':5,
    'nbins':20,
    'learn_rate':0.25,
    'stopping_metric':'mse',
    'score_each_iteration':True,
    'categorical_encoding':'enum',
    'sample_rate':1.0,
    'col_sample_rate':1.0,
    'min_rows':5,
    'distribution':'gaussian'
}

# lightgbm default parameters
lgbm_parameters = {
    'boosting_type':'gbrt',
    'objective':'regression',
    'num_leaves':31,
    'min_data_in_leaf':5,
    'learning_rate':0.25,
    'feature_fraction':1.0,
    'num_iterations':300
}

cat_parameters = {
    'iterations':300,
    'learning_rate':0.25,
    'l2_leaf_reg':3.0,
    'depth':6,
    'has_time':True,
    'bootstrap_type':'No'
}

# fbprophet default parameters 
prophet_kwargs = {
    'n_changepoints':120,
    'changepoint_range':0.9,
    'changepoint_prior_scale':0.05
}
prophet_kwargs_extra = {
    'wk_fourier_order' : 3,
    'wk_prior_scale'   : 10.,
    'yr_fourier_order' : 10,
    'yr_prior_scale'   : 10.
} 

# datatype of sequential calendar features
calendar_sequential_features_types = {
    'week_day':'numeric',
    'month_day':'numeric',
    'year_day':'numeric',
    'year_week':'numeric',
    'month':'categorical',
    'year':'numeric',
    'month_progress':'numeric'
}

# datatype of cyclical calendar features
calendar_cyclical_features_types = {
    'week_day_cos':'numeric',
    'week_day_sin':'numeric',
    'year_day_cos':'numeric',
    'year_day_sin':'numeric',
    'year_week_cos':'numeric',
    'year_week_sin':'numeric',
    'month_cos':'numeric',
    'month_sin':'numeric'
}

# datatype of all features
all_features_types = {
    **calendar_sequential_features_types,
    **calendar_cyclical_features_types
}



import itertools
import pandas as pd
pd.set_option('display.max_colwidth', -1)
from joblib import Parallel, delayed
from tqdm import tqdm


def fit_evaluate(model_class, model_config, model_params, train_data, valid_period, test_data, metric):
    '''
    Parameters
    ----------
    model_class_path: tuple
        Class path of the model: (module_name, class_name)
    '''
    model = model_class(model_params=model_params, **model_config)
    model.fit(train_data, valid_period)
    error = model.evaluate(test_data, metric=metric)
    error = 100*error/test_data.y.mean()
    return (model_params, model.best_iteration, error)

class GridSearch(object):
    '''
    model_class: Class
        class of the model to be instantiated
    features: list
        list of features to be included
    categorical_features: list
        list of names of categorical features
    calendar_anomaly: list
        List of names of calendar features affected by an anomaly
    detrend: bool
        whether or not to remove the trend from time serie
    response_scaling:
        Whether or not to perform scaling of the reponse variable
    lags: list
        List of integer lag values
    window_sizes: list
        List of integer window sizes values
    window_functions: list
        List of string names of the window functions
    hyperparams: dict
        dictionary of hyperparameters in the form: "parameter_name" : [param_value,]
    hyperparams_fixed: dict
        dictionary of fixed hyperparameters in the form: {"param_name" : "param_value",}
    n_jobs: int
        number of parallel jobs to run on grid search
    '''

    def __init__(self, model_class, features=['calendar_mixed','events'], 
                 categorical_features=list(), calendar_anomaly=None, 
                 detrend=True, response_scaling=True, 
                 lags=None, window_sizes=None, window_functions=None, 
                 hyperparams=dict(), hyperparams_fixed=dict(), 
                 n_jobs=-1):
        self.model_class = model_class
        self.features = features
        self.categorical_features = categorical_features
        self.calendar_anomaly = calendar_anomaly
        self.detrend = detrend
        self.response_scaling = response_scaling
        self.lags = lags
        self.window_sizes = window_sizes
        self.window_functions = window_functions
        self.hyperparams = hyperparams
        self.hyperparams_fixed = hyperparams_fixed
        self.n_jobs = n_jobs

    def fit(self, train_data, valid_period=None, test_data=None, metric='rmse'):
        '''
        Parameters
        ----------
        train_data : pandas.DataFrame
            dataframe with at least columns "ds" and "y"
        valid_period: pandas.Series or pandas.DataFrame
            series or dataframe (with column "ds") indicating the validation period
        test_data : pandas.Series or pandas.DataFrame
            dataframe with the same columns as "train_data" over the test period
        metric: string
            name of the error metric to be measured
        Returns
        ----------
        '''
        params_names = self.hyperparams.keys()
        params_values = self.hyperparams.values()
        hyperparams_fixed = self.hyperparams_fixed
        _hyperparams_list = [dict(zip(params_names,combination)) 
                             for combination in itertools.product(*params_values)]
        hyperparams_list = [{**hyperparams, **hyperparams_fixed} 
                            for hyperparams in _hyperparams_list]

        if (valid_period is None) and (test_data is None):
            test_data = train_data.loc[:, ['ds','y']]
        elif test_data is None:
            test_data = pd.merge(train_data, valid_period, how='inner', on=['ds'])

        # parallel fit & evaluation of model on hyperparams
        model_config = {'features':self.features,
                        'categorical_features':self.categorical_features,
                        'calendar_anomaly':self.calendar_anomaly,
                        'detrend':self.detrend,
                        'response_scaling':self.response_scaling,
                        'lags':self.lags,
                        'window_sizes':self.window_sizes,
                        'window_functions':self.window_functions}
        kwargs = {'model_class':self.model_class,
                  'model_config':model_config,
                  'train_data':train_data,
                  'valid_period':valid_period,
                  'test_data':test_data,
                  'metric':metric}
        
        with Parallel(n_jobs=self.n_jobs) as parallel:
            delayed_func = delayed(fit_evaluate)
            with tqdm(hyperparams_list) as tqdm_hyperparams_list:
                _results = parallel(delayed_func(model_params=model_params, **kwargs)
                                    for model_params in tqdm_hyperparams_list)
            tqdm_hyperparams_list.close()
        # removes fixed hyperparams from results
        results = [({key:value for (key,value) in r[0].items() 
                     if key not in hyperparams_fixed.keys()}, r[1], r[2]) 
                    for r in _results]
        # sort the results by error
        results.sort(key = lambda x: x[-1])
        results_dataframe = pd.DataFrame([(r[0],r[1],r[2]) for r in results],
                                columns=["hyperparams","best_iteration","error"])
        self.results = results
        self.results_dataframe = results_dataframe
        
    def get_grid(self, top_n=10):
        '''
        Parameters
        ----------
        top_n: int
            number of best parameters to return
        '''
        return self.results_dataframe.head(top_n)

    def get_best_params(self, top_n=1):
        '''
        Parameters
        ----------
        top_n: int
            number of best parameters to return
        '''
        assert top_n <= len(self.results), \
            '"top_n" cannot be greater that the grid size'
        return [self.results[i][0] for i in range(top_n)]



from zope.interface import Interface

class ForecasterInterface(Interface):
    def __init__(self, model_params, features, detrend):
        pass

    def _prepare_train_data(self, train_data):
        pass
    
    def _prepare_test_data(self, test_data):
        pass

    def _cast_dataframe(self, features_dataframe, features_types):
        pass

    def fit(self, train_data, valid_data=None, early_stopping_rounds=20):
        pass

    def predict(self, test_data):
        pass

    def evaluate(self, test_data, metric='rmse'):
        pass

    def show_variable_importance(self):
        pass

    def save_model(self):
        pass


import warnings
warnings.simplefilter(action='ignore', category=UserWarning)

import numpy as np
import pandas as pd
from zope.interface import implementer
import lightgbm as lgb

from tsforest.config import lgbm_parameters
from tsforest.forecaster_base import ForecasterBase
from tsforest.forecaster_interface import ForecasterInterface


@implementer(ForecasterInterface)
class LightGBMForecaster(ForecasterBase):
    '''
    Parameters
    ----------
    model_params : dict
        Dictionary containing the parameters of the specific boosting model 
    features: list
        List of features to be included
    categorical_features: list
        List of names of categorical features
    calendar_anomaly: list
        List of names of calendar features affected by an anomaly
    detrend: bool
        Whether or not to remove the trend from time series
    response_scaling:
        Whether or not to perform scaling of the reponse variable
    lags: list
        List of integer lag values
    window_sizes: list
        List of integer window sizes values
    window_functions: list
        List of string names of the window functions
    '''
    def __init__(self, model_params=dict(), features=['calendar_mixed'], categorical_features=list(),
                 calendar_anomaly=None, detrend=True, response_scaling=False, 
                 lags=None, window_sizes=None, window_functions=None):

        if lags is not None and 'lag' not in features:
            features.append('lag')
        if (window_sizes is not None and window_functions is not None) and 'rw' not in features:
            features.append('rw')

        self.model_params = model_params
        self.features = features
        self._categorical_features = categorical_features
        self.calendar_anomaly = calendar_anomaly
        self.detrend = detrend
        self.response_scaling = response_scaling
        self.lags = lags
        self.window_sizes = window_sizes
        self.window_functions = window_functions

    def _cast_dataframe(self, features_dataframe, categorical_features):
        """
        Parameters
        ----------
        features_dataframe: pandas.DataFrame
            dataframe containing all the features
        categorical_features: list
            list of names of categorical features
        Returns
        ----------
        features_dataframe_casted: lightgbm.basic.Dataset
            features dataframe casted to LightGBM dataframe format
        """
        dataset_params = {'data':features_dataframe.loc[:, self.input_features],
                          'categorical_feature':categorical_features,
                          'free_raw_data':False}
        if 'weight' in features_dataframe.columns:
            dataset_params['weight'] = features_dataframe.loc[:, 'weight']
        if self.target in features_dataframe.columns:
            dataset_params['label'] = features_dataframe.loc[:, self.target]
        features_dataframe_casted = lgb.Dataset(**dataset_params)
        return features_dataframe_casted     

    def fit(self, train_data, valid_period=None, early_stopping_rounds=20):
        '''
        Parameters
        ----------
        train_data : pandas.DataFrame
            dataframe with "at least" columns "ds" and "y"
        valid_period: pandas.DataFrame
            dataframe (with column "ds") indicating the validation period
        '''
        assert {"ds","y"} <= set(train_data.columns.values), \
            '"train_data" must contain columns "ds" and "y"'
        train_features,categorical_features = super()._prepare_train_features(train_data)

        if valid_period is not None:
            assert (valid_period.ds.min() >= train_data.ds.min() and 
                valid_period.ds.max() <= train_data.ds.max()), \
                'valid_period must be contained in the time period of train_data'
            valid_features = super()._prepare_valid_features(valid_period, train_features)
            valid_start_time = valid_features.ds.min()
            # removes validation period from train data
            train_data = train_data.query('ds < @valid_start_time')
            train_features = train_features.query('ds < @valid_start_time')
        
        train_features['y_hat'] = super()._prepare_train_response(train_features)
        if valid_period is not None:
            valid_features['y_hat'] = super()._prepare_valid_response(valid_features)
        
        train_features_casted = self._cast_dataframe(train_features, categorical_features)
        valid_features_casted = self._cast_dataframe(valid_features, categorical_features) \
                                if valid_period is not None else None

        self.train_data = train_data
        self.categorical_features = categorical_features
        self.train_features = train_features
        self.train_features_casted = train_features_casted

        self.valid_period = valid_period
        self.valid_features = valid_features if valid_period is not None else None
        self.valid_features_casted = valid_features_casted

        # model_params overwrites default params of model
        model_params = {**lgbm_parameters, **self.model_params}

        training_params = {'params':model_params,
                           'train_set':train_features_casted}
        if valid_period is not None:
            training_params['valid_sets'] = valid_features_casted
            training_params['early_stopping_rounds'] = early_stopping_rounds
            training_params['verbose_eval'] = False

        # model training
        model = lgb.train(**training_params)
        self.model = model
        self.best_iteration = model.best_iteration if model.best_iteration>0 else model.num_trees()
    
    def _predict(self, model, test_features, trend_dataframe):
        """
        Parameters
        ----------
        model: LightGBM 
            Trained model
        test_features: pandas.DataFrame
            datafame containing the features for the test period
        trend_dataframe: pandas.DataFrame
            dataframe containing the trend estimation over the test period
        """
        y_train = self.train_features.y.values
        y_valid = self.valid_features.y.values \
                  if self.valid_period is not None else np.array([])
        y = np.concatenate([y_train, y_valid])

        prediction = list()
        for idx in range(test_features.shape[0]):
            if 'lag' in self.features:
                for lag in self.lags:
                    test_features.loc[idx, f'lag_{lag}'] = y[-lag]
            if 'rw' in self.features:
                for window_func in self.window_functions:
                    for window in self.window_sizes:
                        test_features.loc[idx, f'{window_func}_{window}'] = getattr(np, window_func)(y[-window:])
            y_pred = model.predict(test_features.loc[[idx], self.input_features])
            prediction.append(y_pred.copy())
            if self.response_scaling:
                y_pred *= self.y_std
                y_pred += self.y_mean
            if self.detrend:
                y_pred += trend_dataframe.loc[idx, 'trend']
            y = np.append(y, [y_pred])
        return np.asarray(prediction).ravel()

    def predict(self, test_period):
        '''
        Parameters
        ----------
        test_period: pandas.DataFrame
            dataframe with the same columns as "train_data" except for "y"
        Returns
        ----------
        prediction_dataframe: pandas.DataFrame
            dataframe containing dates "ds" and predictions "y_pred"
        '''
        assert set(self.train_data.columns) - set(test_period.columns) == {'y'}, \
            '"test_period" must have the same columns as "train_data" except for "y"'
        
        test_features = super()._prepare_test_features(test_period)
        if self.detrend:
            trend_estimator = self.trend_estimator
            trend_dataframe = trend_estimator.predict(test_period.loc[:, ['ds']])
        else:
            trend_dataframe = None

        if 'lag' in self.features or 'rw' in self.features:
            prediction = self._predict(self.model, test_features, trend_dataframe)
        else:
            prediction = self.model.predict(test_features.loc[:, self.input_features])
        
        if self.response_scaling:
            prediction *= self.y_std
            prediction += self.y_mean
        if self.detrend:
            prediction += trend_dataframe.trend.values
        if 'zero_response' in test_features.columns:
            zero_response_mask = test_features['zero_response']==1
            prediction[zero_response_mask] = 0

        self.test_features = test_features
            
        prediction_dataframe = pd.DataFrame({'ds':test_period.ds, 'y_pred':prediction})
        return prediction_dataframe

    def show_variable_importance(self):
        pass

    def save_model(self):
        pass